--[[
   dynamically create an event catcher frame to do stuff after other addons are
   loaded
  ]]
local MAJOR = "ShortTerm";
local ShortTerm = LibStub("AceAddon-3.0"):NewAddon(MAJOR);
local L = LibStub("AceLocale-3.0"):GetLocale("Short term");
local LG = LibStub("AceLocale-3.0"):GetLocale("Short term message groups");

-- /spew LibStub("AceAddon-3.0"):GetAddon("ShortTerm").cache.data
local LibRingCache = LibStub:GetLibrary("LibRingCache-1.0");
local save = function(self, ...)
	self.cache:append({time(), ...});
end

local events = {
	SYSTEM                = { "CHAT_MSG_SYSTEM","TIME_PLAYED_MSG",
	                          "PLAYER_LEVEL_UP","CHARACTER_POINTS_CHANGED",},
	SAY                   = { "CHAT_MSG_SAY",},
	EMOTE                 = { "CHAT_MSG_EMOTE","CHAT_MSG_TEXT_EMOTE",},
	YELL                  = { "CHAT_MSG_YELL",},
	WHISPER               = { "CHAT_MSG_WHISPER","CHAT_MSG_WHISPER_INFORM",},
	BN_WHISPER            = { "CHAT_MSG_BN_WHISPER","CHAT_MSG_BN_WHISPER_INFORM",},
	BN_CONVERSATION       = { "CHAT_MSG_BN_CONVERSATION",
	                          "CHAT_MSG_BN_CONVERSATION_LIST",
							  "CHAT_MSG_BN_CONVERSATION_NOTICE",},
	BN_TOAST              = { "CHAT_MSG_BN_INLINE_TOAST_ALERT",
	                          "CHAT_MSG_BN_INLINE_TOAST_BROADCAST",
	                          "CHAT_MSG_BN_INLINE_TOAST_BROADCAST_INFORM",
							  "CHAT_MSG_BN_INLINE_TOAST_CONVERSATION",},
	PARTY                 = { "CHAT_MSG_PARTY","CHAT_MSG_PARTY_LEADER",
	                          "CHAT_MSG_MONSTER_PARTY",},
	RAID                  = { "CHAT_MSG_RAID",},
	RAID_LEADER           = { "CHAT_MSG_RAID_LEADER", },
	RAID_WARNING          = { "CHAT_MSG_RAID_WARNING", },
	BATTLEGROUND          = { "CHAT_MSG_BATTLEGROUND", },
	BATTLEGROUND_LEADER   = { "CHAT_MSG_BATTLEGROUND_LEADER", },
	GUILD                 = { "CHAT_MSG_GUILD","GUILD_MOTD", },
	GUILD_OFFICER         = { "CHAT_MSG_OFFICER", },
	MONSTER_SAY           = { "CHAT_MSG_MONSTER_SAY", },
	MONSTER_YELL          = { "CHAT_MSG_MONSTER_YELL", },
	MONSTER_EMOTE         = { "CHAT_MSG_MONSTER_EMOTE", },
	MONSTER_WHISPER       = { "CHAT_MSG_MONSTER_WHISPER", },
	MONSTER_BOSS_EMOTE    = { "CHAT_MSG_RAID_BOSS_EMOTE", },
	MONSTER_BOSS_WHISPER  = { "CHAT_MSG_RAID_BOSS_WHISPER", },
	ERRORS                = { "CHAT_MSG_RESTRICTED","CHAT_MSG_FILTERED", },
	AFK                   = { "CHAT_MSG_AFK", },
	DND                   = { "CHAT_MSG_DND", },
	IGNORED               = { "CHAT_MSG_IGNORED", },
	BG_HORDE              = { "CHAT_MSG_BG_SYSTEM_HORDE", },
	BG_ALLIANCE           = { "CHAT_MSG_BG_SYSTEM_ALLIANCE", },
	BG_NEUTRAL            = { "CHAT_MSG_BG_SYSTEM_NEUTRAL", },
	COMBAT_XP_GAIN        = { "CHAT_MSG_COMBAT_XP_GAIN", },
	COMBAT_HONOR_GAIN     = { "CHAT_MSG_COMBAT_HONOR_GAIN", },
	COMBAT_FACTION_CHANGE = { "CHAT_MSG_COMBAT_FACTION_CHANGE", },
	SKILL                 = { "CHAT_MSG_SKILL", },
	LOOT                  = { "CHAT_MSG_LOOT", },
	MONEY                 = { "CHAT_MSG_MONEY", },
	OPENING               = { "CHAT_MSG_OPENING", },
	TRADESKILLS           = { "CHAT_MSG_TRADESKILLS", },
	PET_INFO              = { "CHAT_MSG_PET_INFO", },
	COMBAT_MISC_INFO      = { "CHAT_MSG_COMBAT_MISC_INFO", },
	ACHIEVEMENT           = { "CHAT_MSG_ACHIEVEMENT", },
	GUILD_ACHIEVEMENT     = { "CHAT_MSG_GUILD_ACHIEVEMENT", },
	CHANNEL               = { "CHAT_MSG_CHANNEL_JOIN",   "CHAT_MSG_CHANNEL_LEAVE",
	                          "CHAT_MSG_CHANNEL_NOTICE", "CHAT_MSG_CHANNEL_NOTICE_USER",
	                          "CHAT_MSG_CHANNEL_LIST",   "CHAT_MSG_CHANNEL", },
};
ShortTerm.Events = events;

-- Cache these values for quick access
local GuildName, CharacterName;
local DBExpanderFrame;

--[[ kReorder(table, num) ]--
 Re-Orders a table by its (numeric) keys. Keys will change their value but not
 their relative position.
 
 This function will not affect non-numeric keys.
 
 This function does not return a table but modify the supplied one.
 
 the optional 2nd parameter will cause this function to discard the first n-num
 items and only leave the remaining num items in the table.
]]
local function kReorder(tbl, num)
	local keys = {};
	for k,v in pairs(tbl) do
		if type(k) == "number" then
			tinsert(keys, k);
		end
	end
	sort(keys);
	
	local start = 1;
	if num and num > 0 and num < #keys then
		start = #keys-num+1;
	end
	
	local j = 1;
	local v;
	-- clear entries 1 to #-num or none if start equals 1
	for i=1, start-1 do
		tbl[keys[i]] = nil;
	end
	for i=start,#keys do
		--print(i, keys[i], "to", j);
		v = tbl[keys[i]];
		tbl[keys[i]] = nil;
		tbl[j] = v
		j = j+1;
	end

	return j;
end

local function RecoverGroup(settings, into)
	local data = settings.Data;
	-- copy data into "into"
	for k, v in pairs(data) do
		tinsert(into, v);
	end
	
	-- create a catcher frame
	local frame = CreateFrame("frame", UIParent);
	frame:Hide();
	frame:SetScript("OnEvent", save);
	
	-- register events
	for k,v in pairs(settings.Events) do
		if v then
			for l,event in pairs(events[k]) do
				frame:RegisterEvent(event);
			end
		end
	end
	
	-- set up ringbuffer
	local cache = LibRingCache:createBlackbox(settings.Length, data);
	frame.cache = cache;
	settings.Frame = nil; -- TODO: remove line before release. This line will be redundant.
	return frame;
end

function ShortTerm:OnEnable()
	-- check for outdated settings/data
	ShortTermMemorySettings = nil; -- obsolete
	if not ShortTermMemory.__VERSION__ or
			ShortTermMemory.__VERSION__ ~=
			self.MemoryVersion then
		ShortTermMemory = self.Defaults.Character;
	end
	if not GlobalShortTermMemory.__VERSION__ or
			GlobalShortTermMemory.__VERSION__ ~=
			self.MemoryVersion then
		GlobalShortTermMemory = self.Defaults.Global;
	end
	
	GuildName = GetGuildInfo("player");
	CharacterName = UnitName("player");
	
	-- this function will overload the global time function
	local savedtime = 0;
	local blizztime = time;
	local blizzdate = date;
	local function ltime()
		return savedtime;
	end
	local function ldate(frmt, t)
		return blizzdate(frmt, t or savedtime);
	end
	_G.time = ltime;
	_G.date = ldate;
	
	-- merge the different caches
	self.Frames = {};
	self.GlobalFrames = {};
	local stm = {};
	for k,v in ipairs(ShortTermMemory) do
		print(MAJOR, v.Name);
		self.Frames[k] = RecoverGroup(v, stm);
	end
	for k,v in ipairs(GlobalShortTermMemory) do
		print(MAJOR, v.Name);
		self.GlobalFrames[k] = RecoverGroup(v, stm);
	end
	
	-- bring the messages into order
	sort(stm, function(a, b) return a[1] < b[1]; end);

	-- load old lines into the chat
	local f;
	local fs = {};
	for k,v in ipairs(stm) do
		savedtime = v[1];
		-- insert into all chat frames
		for i=0, 9 do
			f = _G["ChatFrame"..i];
			if f and f:IsEventRegistered(v[2]) then
				ChatFrame_MessageEventHandler(_G["ChatFrame"..i], select(2, unpack(v)));
				fs[f] = true;
			end
		end
	end
	-- add a line indicating the end of preloaded entries into all chat windows
	-- where lines where added.
	for k in pairs(fs) do
		k:AddMessage(L.MEMORY_END);
	end
	-- reset time function
	_G.time = blizztime;
	_G.date = blizzdate;
end

function ShortTerm:SetUpDBExpander()
	DBExpanderFrame = DBExpanderFrame or CreateFrame("frame");
	
	DBExpanderFrame:RegisterEvent("BN_CONNECTED");
	DBExpanderFrame:SetScript("OnEvent", function(self, e, ...)
		print("Event:", e, "Args:", ...);
	end);
end

function ShortTerm:DisableDBExpander()
	if DBExpanderFrame then
		DBExpanderFrame:UnregisterAllEvents();
	end
end

function ShortTerm:OnDisable()
	return;
end

function ShortTerm:GetGuild(n)
	return GuildName;
end

function ShortTerm:GetCharacter()
	return CharacterName;
end

